@@ -16,24 +16,23 @@ package rules
 import (
 	"fmt"
 	"html/template"
-	"reflect"
 	"sync"
 	"time"
 
-	clientmodel "github.com/prometheus/client_golang/model"
+	"github.com/prometheus/common/model"
 
 	"github.com/prometheus/prometheus/promql"
-	"github.com/prometheus/prometheus/utility"
+	"github.com/prometheus/prometheus/util/strutil"
 )
 
 const (
 	// AlertMetricName is the metric name for synthetic alert timeseries.
-	AlertMetricName clientmodel.LabelValue = "ALERTS"
+	alertMetricName model.LabelValue = "ALERTS"
 
 	// AlertNameLabel is the label name indicating the name of an alert.
-	AlertNameLabel clientmodel.LabelName = "alertname"
+	alertNameLabel model.LabelName = "alertname"
 	// AlertStateLabel is the label name indicating the state of an alert.
-	AlertStateLabel clientmodel.LabelName = "alertstate"
+	alertStateLabel model.LabelName = "alertstate"
 )
 
 // AlertState denotes the state of an active alert.
@@ -41,11 +40,11 @@ type AlertState int
 
 func (s AlertState) String() string {
 	switch s {
-	case Inactive:
+	case StateInactive:
 		return "inactive"
-	case Pending:
+	case StatePending:
 		return "pending"
-	case Firing:
+	case StateFiring:
 		return "firing"
 	default:
 		panic("undefined")
@@ -53,14 +52,14 @@ func (s AlertState) String() string {
 }
 
 const (
-	// Inactive alerts are neither firing nor pending.
-	Inactive AlertState = iota
-	// Pending alerts have been active for less than the configured
-	// threshold duration.
-	Pending
-	// Firing alerts have been active for longer than the configured
-	// threshold duration.
-	Firing
+	// StateInactive is the state of an alert that is either firing nor pending.
+	StateInactive AlertState = iota
+	// StatePending is the state of an alert that has been active for less than
+	// the configured threshold duration.
+	StatePending
+	// StateFiring is the state of an alert that has been active for longer than
+	// the configured threshold duration.
+	StateFiring
 )
 
 // Alert is used to track active (pending/firing) alerts over time.
@@ -68,31 +67,28 @@ type Alert struct {
 	// The name of the alert.
 	Name string
 	// The vector element labelset triggering this alert.
-	Labels clientmodel.LabelSet
+	Labels model.LabelSet
 	// The state of the alert (Pending or Firing).
 	State AlertState
 	// The time when the alert first transitioned into Pending state.
-	ActiveSince clientmodel.Timestamp
+	ActiveSince model.Time
 	// The value of the alert expression for this vector element.
-	Value clientmodel.SampleValue
+	Value model.SampleValue
 }
 
 // sample returns a Sample suitable for recording the alert.
-func (a Alert) sample(timestamp clientmodel.Timestamp, value clientmodel.SampleValue) *promql.Sample {
-	recordedMetric := clientmodel.Metric{}
+func (a Alert) sample(timestamp model.Time, value model.SampleValue) *model.Sample {
+	recordedMetric := make(model.Metric, len(a.Labels)+3)
 	for label, value := range a.Labels {
 		recordedMetric[label] = value
 	}
 
-	recordedMetric[clientmodel.MetricNameLabel] = AlertMetricName
-	recordedMetric[AlertNameLabel] = clientmodel.LabelValue(a.Name)
-	recordedMetric[AlertStateLabel] = clientmodel.LabelValue(a.State.String())
+	recordedMetric[model.MetricNameLabel] = alertMetricName
+	recordedMetric[alertNameLabel] = model.LabelValue(a.Name)
+	recordedMetric[alertStateLabel] = model.LabelValue(a.State.String())
 
-	return &promql.Sample{
-		Metric: clientmodel.COWMetric{
-			Metric: recordedMetric,
-			Copied: true,
-		},
+	return &model.Sample{
+		Metric:    recordedMetric,
 		Value:     value,
 		Timestamp: timestamp,
 	}
@@ -103,22 +99,47 @@ type AlertingRule struct {
 	// The name of the alert.
 	name string
 	// The vector expression from which to generate alerts.
-	Vector promql.Expr
+	vector promql.Expr
 	// The duration for which a labelset needs to persist in the expression
 	// output vector before an alert transitions from Pending to Firing state.
 	holdDuration time.Duration
 	// Extra labels to attach to the resulting alert sample vectors.
-	Labels clientmodel.LabelSet
+	labels model.LabelSet
 	// Short alert summary, suitable for email subjects.
-	Summary string
+	summary string
 	// More detailed alert description.
-	Description string
+	description string
+	// A reference to a runbook for the alert.
+	runbook string
 
 	// Protects the below.
 	mutex sync.Mutex
 	// A map of alerts which are currently active (Pending or Firing), keyed by
 	// the fingerprint of the labelset they correspond to.
-	activeAlerts map[clientmodel.Fingerprint]*Alert
+	activeAlerts map[model.Fingerprint]*Alert
+}
+
+// NewAlertingRule constructs a new AlertingRule.
+func NewAlertingRule(
+	name string,
+	vector promql.Expr,
+	holdDuration time.Duration,
+	labels model.LabelSet,
+	summary string,
+	description string,
+	runbook string,
+) *AlertingRule {
+	return &AlertingRule{
+		name:         name,
+		vector:       vector,
+		holdDuration: holdDuration,
+		labels:       labels,
+		summary:      summary,
+		description:  description,
+		runbook:      runbook,
+
+		activeAlerts: map[model.Fingerprint]*Alert{},
+	}
 }
 
 // Name returns the name of the alert.
@@ -126,19 +147,14 @@ func (rule *AlertingRule) Name() string {
 	return rule.name
 }
 
-// EvalRaw returns the raw value of the rule expression, without creating alerts.
-func (rule *AlertingRule) EvalRaw(timestamp clientmodel.Timestamp, engine *promql.Engine) (promql.Vector, error) {
-	query, err := engine.NewInstantQuery(rule.Vector.String(), timestamp)
+// eval evaluates the rule expression and then creates pending alerts and fires
+// or removes previously pending alerts accordingly.
+func (rule *AlertingRule) eval(timestamp model.Time, engine *promql.Engine) (model.Vector, error) {
+	query, err := engine.NewInstantQuery(rule.vector.String(), timestamp)
 	if err != nil {
 		return nil, err
 	}
-	return query.Exec().Vector()
-}
-
-// Eval evaluates the rule expression and then creates pending alerts and fires
-// or removes previously pending alerts accordingly.
-func (rule *AlertingRule) Eval(timestamp clientmodel.Timestamp, engine *promql.Engine) (promql.Vector, error) {
-	exprResult, err := rule.EvalRaw(timestamp, engine)
+	exprResult, err := query.Exec().Vector()
 	if err != nil {
 		return nil, err
 	}
@@ -148,22 +164,21 @@ func (rule *AlertingRule) Eval(timestamp clientmodel.Timestamp, engine *promql.E
 
 	// Create pending alerts for any new vector elements in the alert expression
 	// or update the expression value for existing elements.
-	resultFingerprints := utility.Set{}
+	resultFPs := map[model.Fingerprint]struct{}{}
 	for _, sample := range exprResult {
-		fp := sample.Metric.Metric.Fingerprint()
-		resultFingerprints.Add(fp)
+		fp := sample.Metric.Fingerprint()
+		resultFPs[fp] = struct{}{}
 
 		if alert, ok := rule.activeAlerts[fp]; !ok {
-			labels := clientmodel.LabelSet{}
-			labels.MergeFromMetric(sample.Metric.Metric)
-			labels = labels.Merge(rule.Labels)
-			if _, ok := labels[clientmodel.MetricNameLabel]; ok {
-				delete(labels, clientmodel.MetricNameLabel)
+			labels := model.LabelSet(sample.Metric.Clone())
+			labels = labels.Merge(rule.labels)
+			if _, ok := labels[model.MetricNameLabel]; ok {
+				delete(labels, model.MetricNameLabel)
 			}
 			rule.activeAlerts[fp] = &Alert{
 				Name:        rule.name,
 				Labels:      labels,
-				State:       Pending,
+				State:       StatePending,
 				ActiveSince: timestamp,
 				Value:       sample.Value,
 			}
@@ -172,19 +187,19 @@ func (rule *AlertingRule) Eval(timestamp clientmodel.Timestamp, engine *promql.E
 		}
 	}
 
-	vector := promql.Vector{}
+	var vector model.Vector
 
 	// Check if any pending alerts should be removed or fire now. Write out alert timeseries.
 	for fp, activeAlert := range rule.activeAlerts {
-		if !resultFingerprints.Has(fp) {
+		if _, ok := resultFPs[fp]; !ok {
 			vector = append(vector, activeAlert.sample(timestamp, 0))
 			delete(rule.activeAlerts, fp)
 			continue
 		}
 
-		if activeAlert.State == Pending && timestamp.Sub(activeAlert.ActiveSince) >= rule.holdDuration {
+		if activeAlert.State == StatePending && timestamp.Sub(activeAlert.ActiveSince) >= rule.holdDuration {
 			vector = append(vector, activeAlert.sample(timestamp, 0))
-			activeAlert.State = Firing
+			activeAlert.State = StateFiring
 		}
 
 		vector = append(vector, activeAlert.sample(timestamp, 1))
@@ -193,39 +208,41 @@ func (rule *AlertingRule) Eval(timestamp clientmodel.Timestamp, engine *promql.E
 	return vector, nil
 }
 
-// DotGraph returns the text representation of a dot graph.
-func (rule *AlertingRule) DotGraph() string {
-	graph := fmt.Sprintf(
-		`digraph "Rules" {
-	  %#p[shape="box",label="ALERT %s IF FOR %s"];
-		%#p -> %x;
-		%s
-	}`,
-		&rule, rule.name, utility.DurationToString(rule.holdDuration),
-		&rule, reflect.ValueOf(rule.Vector).Pointer(),
-		rule.Vector.DotGraph(),
-	)
-	return graph
-}
-
 func (rule *AlertingRule) String() string {
-	return fmt.Sprintf("ALERT %s IF %s FOR %s WITH %s", rule.name, rule.Vector, utility.DurationToString(rule.holdDuration), rule.Labels)
+	s := fmt.Sprintf("ALERT %s", rule.name)
+	s += fmt.Sprintf("\n\tIF %s", rule.vector)
+	if rule.holdDuration > 0 {
+		s += fmt.Sprintf("\n\tFOR %s", strutil.DurationToString(rule.holdDuration))
+	}
+	if len(rule.labels) > 0 {
+		s += fmt.Sprintf("\n\tWITH %s", rule.labels)
+	}
+	s += fmt.Sprintf("\n\tSUMMARY %q", rule.summary)
+	s += fmt.Sprintf("\n\tDESCRIPTION %q", rule.description)
+	s += fmt.Sprintf("\n\tRUNBOOK %q", rule.runbook)
+	return s
 }
 
-// HTMLSnippet returns an HTML snippet representing this alerting rule.
+// HTMLSnippet returns an HTML snippet representing this alerting rule. The
+// resulting snippet is expected to be presented in a <pre> element, so that
+// line breaks and other returned whitespace is respected.
 func (rule *AlertingRule) HTMLSnippet(pathPrefix string) template.HTML {
-	alertMetric := clientmodel.Metric{
-		clientmodel.MetricNameLabel: AlertMetricName,
-		AlertNameLabel:              clientmodel.LabelValue(rule.name),
+	alertMetric := model.Metric{
+		model.MetricNameLabel: alertMetricName,
+		alertNameLabel:        model.LabelValue(rule.name),
 	}
-	return template.HTML(fmt.Sprintf(
-		`ALERT <a href="%s">%s</a> IF <a href="%s">%s</a> FOR %s WITH %s`,
-		pathPrefix+utility.GraphLinkForExpression(alertMetric.String()),
-		rule.name,
-		pathPrefix+utility.GraphLinkForExpression(rule.Vector.String()),
-		rule.Vector,
-		utility.DurationToString(rule.holdDuration),
-		rule.Labels))
+	s := fmt.Sprintf("ALERT <a href=%q>%s</a>", pathPrefix+strutil.GraphLinkForExpression(alertMetric.String()), rule.name)
+	s += fmt.Sprintf("\n  IF <a href=%q>%s</a>", pathPrefix+strutil.GraphLinkForExpression(rule.vector.String()), rule.vector)
+	if rule.holdDuration > 0 {
+		s += fmt.Sprintf("\n  FOR %s", strutil.DurationToString(rule.holdDuration))
+	}
+	if len(rule.labels) > 0 {
+		s += fmt.Sprintf("\n  WITH %s", rule.labels)
+	}
+	s += fmt.Sprintf("\n  SUMMARY %q", rule.summary)
+	s += fmt.Sprintf("\n  DESCRIPTION %q", rule.description)
+	s += fmt.Sprintf("\n  RUNBOOK %q", rule.runbook)
+	return template.HTML(s)
 }
 
 // State returns the "maximum" state: firing > pending > inactive.
@@ -233,7 +250,7 @@ func (rule *AlertingRule) State() AlertState {
 	rule.mutex.Lock()
 	defer rule.mutex.Unlock()
 
-	maxState := Inactive
+	maxState := StateInactive
 	for _, activeAlert := range rule.activeAlerts {
 		if activeAlert.State > maxState {
 			maxState = activeAlert.State
@@ -253,17 +270,3 @@ func (rule *AlertingRule) ActiveAlerts() []Alert {
 	}
 	return alerts
 }
-
-// NewAlertingRule constructs a new AlertingRule.
-func NewAlertingRule(name string, vector promql.Expr, holdDuration time.Duration, labels clientmodel.LabelSet, summary string, description string) *AlertingRule {
-	return &AlertingRule{
-		name:         name,
-		Vector:       vector,
-		holdDuration: holdDuration,
-		Labels:       labels,
-		Summary:      summary,
-		Description:  description,
-
-		activeAlerts: map[clientmodel.Fingerprint]*Alert{},
-	}
-}
