@@ -16,21 +16,23 @@ package rules
 import (
 	"fmt"
 	"io/ioutil"
-	"strings"
+	"net/url"
+	"path/filepath"
 	"sync"
 	"time"
 
-	"github.com/prometheus/client_golang/prometheus"
-	"github.com/prometheus/log"
+	html_template "html/template"
 
-	clientmodel "github.com/prometheus/client_golang/model"
+	"github.com/prometheus/client_golang/prometheus"
+	"github.com/prometheus/common/log"
+	"github.com/prometheus/common/model"
 
 	"github.com/prometheus/prometheus/config"
 	"github.com/prometheus/prometheus/notification"
 	"github.com/prometheus/prometheus/promql"
 	"github.com/prometheus/prometheus/storage"
-	"github.com/prometheus/prometheus/templates"
-	"github.com/prometheus/prometheus/utility"
+	"github.com/prometheus/prometheus/template"
+	"github.com/prometheus/prometheus/util/strutil"
 )
 
 // Constants for instrumentation.
@@ -38,8 +40,8 @@ const (
 	namespace = "prometheus"
 
 	ruleTypeLabel     = "rule_type"
-	alertingRuleType  = "alerting"
-	recordingRuleType = "recording"
+	ruleTypeAlerting  = "alerting"
+	ruleTypeRecording = "recording"
 )
 
 var (
@@ -72,6 +74,20 @@ func init() {
 	prometheus.MustRegister(evalDuration)
 }
 
+// A Rule encapsulates a vector expression which is evaluated at a specified
+// interval and acted upon (currently either recorded or used for alerting).
+type Rule interface {
+	// Name returns the name of the rule.
+	Name() string
+	// Eval evaluates the rule, including any associated recording or alerting actions.
+	eval(model.Time, *promql.Engine) (model.Vector, error)
+	// String returns a human-readable string representation of the rule.
+	String() string
+	// HTMLSnippet returns a human-readable string representation of the rule,
+	// decorated with HTML elements for use the web frontend.
+	HTMLSnippet(pathPrefix string) html_template.HTML
+}
+
 // The Manager manages recording and alerting rules.
 type Manager struct {
 	// Protects the rules list.
@@ -86,8 +102,7 @@ type Manager struct {
 	sampleAppender      storage.SampleAppender
 	notificationHandler *notification.NotificationHandler
 
-	prometheusURL string
-	pathPrefix    string
+	externalURL *url.URL
 }
 
 // ManagerOptions bundles options for the Manager.
@@ -98,8 +113,7 @@ type ManagerOptions struct {
 	NotificationHandler *notification.NotificationHandler
 	SampleAppender      storage.SampleAppender
 
-	PrometheusURL string
-	PathPrefix    string
+	ExternalURL *url.URL
 }
 
 // NewManager returns an implementation of Manager, ready to be started
@@ -113,7 +127,7 @@ func NewManager(o *ManagerOptions) *Manager {
 		sampleAppender:      o.SampleAppender,
 		queryEngine:         o.QueryEngine,
 		notificationHandler: o.NotificationHandler,
-		prometheusURL:       o.PrometheusURL,
+		externalURL:         o.ExternalURL,
 	}
 	return manager
 }
@@ -164,7 +178,7 @@ func (m *Manager) Stop() {
 	m.done <- true
 }
 
-func (m *Manager) queueAlertNotifications(rule *AlertingRule, timestamp clientmodel.Timestamp) {
+func (m *Manager) queueAlertNotifications(rule *AlertingRule, timestamp model.Time) {
 	activeAlerts := rule.ActiveAlerts()
 	if len(activeAlerts) == 0 {
 		return
@@ -172,7 +186,7 @@ func (m *Manager) queueAlertNotifications(rule *AlertingRule, timestamp clientmo
 
 	notifications := make(notification.NotificationReqs, 0, len(activeAlerts))
 	for _, aa := range activeAlerts {
-		if aa.State != Firing {
+		if aa.State != StateFiring {
 			// BUG: In the future, make AlertManager support pending alerts?
 			continue
 		}
@@ -184,18 +198,18 @@ func (m *Manager) queueAlertNotifications(rule *AlertingRule, timestamp clientmo
 		}
 		tmplData := struct {
 			Labels map[string]string
-			Value  clientmodel.SampleValue
+			Value  float64
 		}{
 			Labels: l,
-			Value:  aa.Value,
+			Value:  float64(aa.Value),
 		}
 		// Inject some convenience variables that are easier to remember for users
 		// who are not used to Go's templating system.
 		defs := "{{$labels := .Labels}}{{$value := .Value}}"
 
 		expand := func(text string) string {
-			template := templates.NewTemplateExpander(defs+text, "__alert_"+rule.Name(), tmplData, timestamp, m.queryEngine, m.pathPrefix)
-			result, err := template.Expand()
+			tmpl := template.NewTemplateExpander(defs+text, "__alert_"+rule.Name(), tmplData, timestamp, m.queryEngine, m.externalURL.Path)
+			result, err := tmpl.Expand()
 			if err != nil {
 				result = err.Error()
 				log.Warnf("Error expanding alert template %v with data '%v': %v", rule.Name(), tmplData, err)
@@ -204,22 +218,23 @@ func (m *Manager) queueAlertNotifications(rule *AlertingRule, timestamp clientmo
 		}
 
 		notifications = append(notifications, &notification.NotificationReq{
-			Summary:     expand(rule.Summary),
-			Description: expand(rule.Description),
-			Labels: aa.Labels.Merge(clientmodel.LabelSet{
-				AlertNameLabel: clientmodel.LabelValue(rule.Name()),
+			Summary:     expand(rule.summary),
+			Description: expand(rule.description),
+			Runbook:     rule.runbook,
+			Labels: aa.Labels.Merge(model.LabelSet{
+				alertNameLabel: model.LabelValue(rule.Name()),
 			}),
 			Value:        aa.Value,
 			ActiveSince:  aa.ActiveSince.Time(),
 			RuleString:   rule.String(),
-			GeneratorURL: m.prometheusURL + strings.TrimLeft(utility.GraphLinkForExpression(rule.Vector.String()), "/"),
+			GeneratorURL: m.externalURL.String() + strutil.GraphLinkForExpression(rule.vector.String()),
 		})
 	}
 	m.notificationHandler.SubmitReqs(notifications)
 }
 
 func (m *Manager) runIteration() {
-	now := clientmodel.Now()
+	now := model.Now()
 	wg := sync.WaitGroup{}
 
 	m.Lock()
@@ -234,7 +249,7 @@ func (m *Manager) runIteration() {
 			defer wg.Done()
 
 			start := time.Now()
-			vector, err := rule.Eval(now, m.queryEngine)
+			vector, err := rule.eval(now, m.queryEngine)
 			duration := time.Since(start)
 
 			if err != nil {
@@ -246,46 +261,83 @@ func (m *Manager) runIteration() {
 			switch r := rule.(type) {
 			case *AlertingRule:
 				m.queueAlertNotifications(r, now)
-				evalDuration.WithLabelValues(alertingRuleType).Observe(
+				evalDuration.WithLabelValues(ruleTypeAlerting).Observe(
 					float64(duration / time.Millisecond),
 				)
 			case *RecordingRule:
-				evalDuration.WithLabelValues(recordingRuleType).Observe(
+				evalDuration.WithLabelValues(ruleTypeRecording).Observe(
 					float64(duration / time.Millisecond),
 				)
 			default:
-				panic(fmt.Errorf("Unknown rule type: %T", rule))
+				panic(fmt.Errorf("unknown rule type: %T", rule))
 			}
 
 			for _, s := range vector {
-				m.sampleAppender.Append(&clientmodel.Sample{
-					Metric:    s.Metric.Metric,
-					Value:     s.Value,
-					Timestamp: s.Timestamp,
-				})
+				m.sampleAppender.Append(s)
 			}
 		}(rule)
 	}
 	wg.Wait()
 }
 
+// transferAlertState makes a copy of the state of alerting rules and returns a function
+// that restores them in the current state.
+func (m *Manager) transferAlertState() func() {
+
+	alertingRules := map[string]*AlertingRule{}
+	for _, r := range m.rules {
+		if ar, ok := r.(*AlertingRule); ok {
+			alertingRules[ar.name] = ar
+		}
+	}
+
+	return func() {
+		// Restore alerting rule state.
+		for _, r := range m.rules {
+			ar, ok := r.(*AlertingRule)
+			if !ok {
+				continue
+			}
+			if old, ok := alertingRules[ar.name]; ok {
+				ar.activeAlerts = old.activeAlerts
+			}
+		}
+	}
+}
+
 // ApplyConfig updates the rule manager's state as the config requires. If
-// loading the new rules failed the old rule set is restored.
-func (m *Manager) ApplyConfig(conf *config.Config) {
+// loading the new rules failed the old rule set is restored. Returns true on success.
+func (m *Manager) ApplyConfig(conf *config.Config) bool {
 	m.Lock()
 	defer m.Unlock()
 
+	defer m.transferAlertState()()
+
+	success := true
 	m.interval = time.Duration(conf.GlobalConfig.EvaluationInterval)
 
 	rulesSnapshot := make([]Rule, len(m.rules))
 	copy(rulesSnapshot, m.rules)
 	m.rules = m.rules[:0]
 
-	if err := m.loadRuleFiles(conf.RuleFiles...); err != nil {
+	var files []string
+	for _, pat := range conf.RuleFiles {
+		fs, err := filepath.Glob(pat)
+		if err != nil {
+			// The only error can be a bad pattern.
+			log.Errorf("Error retrieving rule files for %s: %s", pat, err)
+			success = false
+		}
+		files = append(files, fs...)
+	}
+	if err := m.loadRuleFiles(files...); err != nil {
 		// If loading the new rules failed, restore the old rule set.
 		m.rules = rulesSnapshot
 		log.Errorf("Error loading rules, previous rule set restored: %s", err)
+		success = false
 	}
+
+	return success
 }
 
 // loadRuleFiles loads alerting and recording rules from the given files.
@@ -299,13 +351,14 @@ func (m *Manager) loadRuleFiles(filenames ...string) error {
 		if err != nil {
 			return fmt.Errorf("error parsing %s: %s", fn, err)
 		}
+
 		for _, stmt := range stmts {
 			switch r := stmt.(type) {
 			case *promql.AlertStmt:
-				rule := NewAlertingRule(r.Name, r.Expr, r.Duration, r.Labels, r.Summary, r.Description)
+				rule := NewAlertingRule(r.Name, r.Expr, r.Duration, r.Labels, r.Summary, r.Description, r.Runbook)
 				m.rules = append(m.rules, rule)
 			case *promql.RecordStmt:
-				rule := &RecordingRule{r.Name, r.Expr, r.Labels}
+				rule := NewRecordingRule(r.Name, r.Expr, r.Labels)
 				m.rules = append(m.rules, rule)
 			default:
 				panic("retrieval.Manager.LoadRuleFiles: unknown statement type")
