@@ -16,48 +16,76 @@ package rules
 import (
 	"fmt"
 	"html/template"
-	"reflect"
 
-	clientmodel "github.com/prometheus/client_golang/model"
+	"github.com/prometheus/common/model"
 
 	"github.com/prometheus/prometheus/promql"
-	"github.com/prometheus/prometheus/utility"
+	"github.com/prometheus/prometheus/util/strutil"
 )
 
 // A RecordingRule records its vector expression into new timeseries.
 type RecordingRule struct {
 	name   string
 	vector promql.Expr
-	labels clientmodel.LabelSet
+	labels model.LabelSet
+}
+
+// NewRecordingRule returns a new recording rule.
+func NewRecordingRule(name string, vector promql.Expr, labels model.LabelSet) *RecordingRule {
+	return &RecordingRule{
+		name:   name,
+		vector: vector,
+		labels: labels,
+	}
 }
 
 // Name returns the rule name.
 func (rule RecordingRule) Name() string { return rule.name }
 
-// EvalRaw returns the raw value of the rule expression.
-func (rule RecordingRule) EvalRaw(timestamp clientmodel.Timestamp, engine *promql.Engine) (promql.Vector, error) {
+// eval evaluates the rule and then overrides the metric names and labels accordingly.
+func (rule RecordingRule) eval(timestamp model.Time, engine *promql.Engine) (model.Vector, error) {
 	query, err := engine.NewInstantQuery(rule.vector.String(), timestamp)
 	if err != nil {
 		return nil, err
 	}
-	return query.Exec().Vector()
-}
 
-// Eval evaluates the rule and then overrides the metric names and labels accordingly.
-func (rule RecordingRule) Eval(timestamp clientmodel.Timestamp, engine *promql.Engine) (promql.Vector, error) {
-	vector, err := rule.EvalRaw(timestamp, engine)
-	if err != nil {
+	var (
+		result = query.Exec()
+		vector model.Vector
+	)
+	if result.Err != nil {
 		return nil, err
 	}
 
+	switch result.Value.(type) {
+	case model.Vector:
+		vector, err = result.Vector()
+		if err != nil {
+			return nil, err
+		}
+	case *model.Scalar:
+		scalar, err := result.Scalar()
+		if err != nil {
+			return nil, err
+		}
+		vector = model.Vector{&model.Sample{
+			Value:     scalar.Value,
+			Timestamp: scalar.Timestamp,
+			Metric:    model.Metric{},
+		}}
+	default:
+		return nil, fmt.Errorf("rule result is not a vector or scalar")
+	}
+
 	// Override the metric name and labels.
 	for _, sample := range vector {
-		sample.Metric.Set(clientmodel.MetricNameLabel, clientmodel.LabelValue(rule.name))
+		sample.Metric[model.MetricNameLabel] = model.LabelValue(rule.name)
+
 		for label, value := range rule.labels {
 			if value == "" {
-				sample.Metric.Delete(label)
+				delete(sample.Metric, label)
 			} else {
-				sample.Metric.Set(label, value)
+				sample.Metric[label] = value
 			}
 		}
 	}
@@ -65,21 +93,6 @@ func (rule RecordingRule) Eval(timestamp clientmodel.Timestamp, engine *promql.E
 	return vector, nil
 }
 
-// DotGraph returns the text representation of a dot graph.
-func (rule RecordingRule) DotGraph() string {
-	graph := fmt.Sprintf(
-		`digraph "Rules" {
-	  %#p[shape="box",label="%s = "];
-		%#p -> %x;
-		%s
-	}`,
-		&rule, rule.name,
-		&rule, reflect.ValueOf(rule.vector).Pointer(),
-		rule.vector.DotGraph(),
-	)
-	return graph
-}
-
 func (rule RecordingRule) String() string {
 	return fmt.Sprintf("%s%s = %s\n", rule.name, rule.labels, rule.vector)
 }
@@ -89,9 +102,9 @@ func (rule RecordingRule) HTMLSnippet(pathPrefix string) template.HTML {
 	ruleExpr := rule.vector.String()
 	return template.HTML(fmt.Sprintf(
 		`<a href="%s">%s</a>%s = <a href="%s">%s</a>`,
-		pathPrefix+utility.GraphLinkForExpression(rule.name),
+		pathPrefix+strutil.GraphLinkForExpression(rule.name),
 		rule.name,
 		rule.labels,
-		pathPrefix+utility.GraphLinkForExpression(ruleExpr),
+		pathPrefix+strutil.GraphLinkForExpression(ruleExpr),
 		ruleExpr))
 }
