@@ -17,52 +17,52 @@ import (
 	"reflect"
 	"sync"
 	"testing"
+	"time"
 
-	clientmodel "github.com/prometheus/client_golang/model"
+	"github.com/prometheus/common/model"
 
 	"github.com/prometheus/prometheus/storage/local/codable"
 	"github.com/prometheus/prometheus/storage/local/index"
-	"github.com/prometheus/prometheus/storage/metric"
-	"github.com/prometheus/prometheus/utility/test"
+	"github.com/prometheus/prometheus/util/testutil"
 )
 
 var (
-	m1 = clientmodel.Metric{"label": "value1"}
-	m2 = clientmodel.Metric{"label": "value2"}
-	m3 = clientmodel.Metric{"label": "value3"}
-	m4 = clientmodel.Metric{"label": "value4"}
-	m5 = clientmodel.Metric{"label": "value5"}
+	m1 = model.Metric{"label": "value1"}
+	m2 = model.Metric{"label": "value2"}
+	m3 = model.Metric{"label": "value3"}
+	m4 = model.Metric{"label": "value4"}
+	m5 = model.Metric{"label": "value5"}
 )
 
-func newTestPersistence(t *testing.T, encoding chunkEncoding) (*persistence, test.Closer) {
-	*defaultChunkEncoding = int(encoding)
-	dir := test.NewTemporaryDirectory("test_persistence", t)
-	p, err := newPersistence(dir.Path(), false, false, func() bool { return false })
+func newTestPersistence(t *testing.T, encoding chunkEncoding) (*persistence, testutil.Closer) {
+	DefaultChunkEncoding = encoding
+	dir := testutil.NewTemporaryDirectory("test_persistence", t)
+	p, err := newPersistence(dir.Path(), false, false, func() bool { return false }, 0.1)
 	if err != nil {
 		dir.Close()
 		t.Fatal(err)
 	}
 	go p.run()
-	return p, test.NewCallbackCloser(func() {
+	return p, testutil.NewCallbackCloser(func() {
 		p.close()
 		dir.Close()
 	})
 }
 
-func buildTestChunks(encoding chunkEncoding) map[clientmodel.Fingerprint][]chunk {
-	fps := clientmodel.Fingerprints{
+func buildTestChunks(encoding chunkEncoding) map[model.Fingerprint][]chunk {
+	fps := model.Fingerprints{
 		m1.FastFingerprint(),
 		m2.FastFingerprint(),
 		m3.FastFingerprint(),
 	}
-	fpToChunks := map[clientmodel.Fingerprint][]chunk{}
+	fpToChunks := map[model.Fingerprint][]chunk{}
 
 	for _, fp := range fps {
 		fpToChunks[fp] = make([]chunk, 0, 10)
 		for i := 0; i < 10; i++ {
-			fpToChunks[fp] = append(fpToChunks[fp], newChunkForEncoding(encoding).add(&metric.SamplePair{
-				Timestamp: clientmodel.Timestamp(i),
-				Value:     clientmodel.SampleValue(fp),
+			fpToChunks[fp] = append(fpToChunks[fp], newChunkForEncoding(encoding).add(&model.SamplePair{
+				Timestamp: model.Time(i),
+				Value:     model.SampleValue(fp),
 			})[0])
 		}
 	}
@@ -88,11 +88,11 @@ func testPersistLoadDropChunks(t *testing.T, encoding chunkEncoding) {
 
 	for fp, chunks := range fpToChunks {
 		firstTimeNotDropped, offset, numDropped, allDropped, err :=
-			p.dropAndPersistChunks(fp, clientmodel.Earliest, chunks)
+			p.dropAndPersistChunks(fp, model.Earliest, chunks)
 		if err != nil {
 			t.Fatal(err)
 		}
-		if got, want := firstTimeNotDropped, clientmodel.Timestamp(0); got != want {
+		if got, want := firstTimeNotDropped, model.Time(0); got != want {
 			t.Errorf("Want firstTimeNotDropped %v, got %v.", got, want)
 		}
 		if got, want := offset, 0; got != want {
@@ -121,12 +121,12 @@ func testPersistLoadDropChunks(t *testing.T, encoding chunkEncoding) {
 			}
 		}
 		// Load all chunk descs.
-		actualChunkDescs, err := p.loadChunkDescs(fp, 10)
+		actualChunkDescs, err := p.loadChunkDescs(fp, 0)
 		if len(actualChunkDescs) != 10 {
 			t.Errorf("Got %d chunkDescs, want %d.", len(actualChunkDescs), 10)
 		}
 		for i, cd := range actualChunkDescs {
-			if cd.firstTime() != clientmodel.Timestamp(i) || cd.lastTime() != clientmodel.Timestamp(i) {
+			if cd.firstTime() != model.Time(i) || cd.lastTime() != model.Time(i) {
 				t.Errorf(
 					"Want ts=%v, got firstTime=%v, lastTime=%v.",
 					i, cd.firstTime(), cd.lastTime(),
@@ -140,7 +140,7 @@ func testPersistLoadDropChunks(t *testing.T, encoding chunkEncoding) {
 			t.Errorf("Got %d chunkDescs, want %d.", len(actualChunkDescs), 5)
 		}
 		for i, cd := range actualChunkDescs {
-			if cd.firstTime() != clientmodel.Timestamp(i) || cd.lastTime() != clientmodel.Timestamp(i) {
+			if cd.firstTime() != model.Time(i) || cd.lastTime() != model.Time(i) {
 				t.Errorf(
 					"Want ts=%v, got firstTime=%v, lastTime=%v.",
 					i, cd.firstTime(), cd.lastTime(),
@@ -203,11 +203,11 @@ func testPersistLoadDropChunks(t *testing.T, encoding chunkEncoding) {
 	// Re-add first two of the chunks.
 	for fp, chunks := range fpToChunks {
 		firstTimeNotDropped, offset, numDropped, allDropped, err :=
-			p.dropAndPersistChunks(fp, clientmodel.Earliest, chunks[:2])
+			p.dropAndPersistChunks(fp, model.Earliest, chunks[:2])
 		if err != nil {
 			t.Fatal(err)
 		}
-		if got, want := firstTimeNotDropped, clientmodel.Timestamp(0); got != want {
+		if got, want := firstTimeNotDropped, model.Time(0); got != want {
 			t.Errorf("Want firstTimeNotDropped %v, got %v.", got, want)
 		}
 		if got, want := offset, 0; got != want {
@@ -338,6 +338,85 @@ func testPersistLoadDropChunks(t *testing.T, encoding chunkEncoding) {
 			t.Error("not all chunks dropped")
 		}
 	}
+	// Now set minShrinkRatio to 0.25 and play with it.
+	p.minShrinkRatio = 0.25
+	// Re-add 8 chunks.
+	for fp, chunks := range fpToChunks {
+		firstTimeNotDropped, offset, numDropped, allDropped, err :=
+			p.dropAndPersistChunks(fp, model.Earliest, chunks[:8])
+		if err != nil {
+			t.Fatal(err)
+		}
+		if got, want := firstTimeNotDropped, model.Time(0); got != want {
+			t.Errorf("Want firstTimeNotDropped %v, got %v.", got, want)
+		}
+		if got, want := offset, 0; got != want {
+			t.Errorf("Want offset %v, got %v.", got, want)
+		}
+		if got, want := numDropped, 0; got != want {
+			t.Errorf("Want numDropped %v, got %v.", got, want)
+		}
+		if allDropped {
+			t.Error("All dropped.")
+		}
+	}
+	// Drop only the first chunk should not happen, but persistence should still work.
+	for fp, chunks := range fpToChunks {
+		firstTime, offset, numDropped, allDropped, err := p.dropAndPersistChunks(fp, 1, chunks[8:9])
+		if err != nil {
+			t.Fatal(err)
+		}
+		if offset != 8 {
+			t.Errorf("want offset 8, got %d", offset)
+		}
+		if firstTime != 0 {
+			t.Errorf("want first time 0, got %d", firstTime)
+		}
+		if numDropped != 0 {
+			t.Errorf("want 0 dropped chunk, got %v", numDropped)
+		}
+		if allDropped {
+			t.Error("all chunks dropped")
+		}
+	}
+	// Drop only the first two chunks should not happen, either.
+	for fp := range fpToChunks {
+		firstTime, offset, numDropped, allDropped, err := p.dropAndPersistChunks(fp, 2, nil)
+		if err != nil {
+			t.Fatal(err)
+		}
+		if offset != 0 {
+			t.Errorf("want offset 0, got %d", offset)
+		}
+		if firstTime != 0 {
+			t.Errorf("want first time 0, got %d", firstTime)
+		}
+		if numDropped != 0 {
+			t.Errorf("want 0 dropped chunk, got %v", numDropped)
+		}
+		if allDropped {
+			t.Error("all chunks dropped")
+		}
+	}
+	// Drop the first three chunks should finally work.
+	for fp, chunks := range fpToChunks {
+		firstTime, offset, numDropped, allDropped, err := p.dropAndPersistChunks(fp, 3, chunks[9:])
+		if err != nil {
+			t.Fatal(err)
+		}
+		if offset != 6 {
+			t.Errorf("want offset 6, got %d", offset)
+		}
+		if firstTime != 3 {
+			t.Errorf("want first time 3, got %d", firstTime)
+		}
+		if numDropped != 3 {
+			t.Errorf("want 3 dropped chunk, got %v", numDropped)
+		}
+		if allDropped {
+			t.Error("all chunks dropped")
+		}
+	}
 }
 
 func TestPersistLoadDropChunksType0(t *testing.T) {
@@ -354,23 +433,23 @@ func testCheckpointAndLoadSeriesMapAndHeads(t *testing.T, encoding chunkEncoding
 
 	fpLocker := newFingerprintLocker(10)
 	sm := newSeriesMap()
-	s1 := newMemorySeries(m1, true, 0)
-	s2 := newMemorySeries(m2, false, 0)
-	s3 := newMemorySeries(m3, false, 0)
-	s4 := newMemorySeries(m4, true, 0)
-	s5 := newMemorySeries(m5, true, 0)
-	s1.add(&metric.SamplePair{Timestamp: 1, Value: 3.14})
-	s3.add(&metric.SamplePair{Timestamp: 2, Value: 2.7})
+	s1 := newMemorySeries(m1, nil, time.Time{})
+	s2 := newMemorySeries(m2, nil, time.Time{})
+	s3 := newMemorySeries(m3, nil, time.Time{})
+	s4 := newMemorySeries(m4, nil, time.Time{})
+	s5 := newMemorySeries(m5, nil, time.Time{})
+	s1.add(&model.SamplePair{Timestamp: 1, Value: 3.14})
+	s3.add(&model.SamplePair{Timestamp: 2, Value: 2.7})
 	s3.headChunkClosed = true
 	s3.persistWatermark = 1
 	for i := 0; i < 10000; i++ {
-		s4.add(&metric.SamplePair{
-			Timestamp: clientmodel.Timestamp(i),
-			Value:     clientmodel.SampleValue(i) / 2,
+		s4.add(&model.SamplePair{
+			Timestamp: model.Time(i),
+			Value:     model.SampleValue(i) / 2,
 		})
-		s5.add(&metric.SamplePair{
-			Timestamp: clientmodel.Timestamp(i),
-			Value:     clientmodel.SampleValue(i * i),
+		s5.add(&model.SamplePair{
+			Timestamp: model.Time(i),
+			Value:     model.SampleValue(i * i),
 		})
 	}
 	s5.persistWatermark = 3
@@ -416,8 +495,8 @@ func testCheckpointAndLoadSeriesMapAndHeads(t *testing.T, encoding chunkEncoding
 		if loadedS3.head().c != nil {
 			t.Error("head chunk not evicted")
 		}
-		if loadedS3.chunkDescsOffset != -1 {
-			t.Errorf("want chunkDescsOffset -1, got %d", loadedS3.chunkDescsOffset)
+		if loadedS3.chunkDescsOffset != 0 {
+			t.Errorf("want chunkDescsOffset 0, got %d", loadedS3.chunkDescsOffset)
 		}
 		if !loadedS3.headChunkClosed {
 			t.Error("headChunkClosed is false")
@@ -490,11 +569,11 @@ func TestCheckpointAndLoadFPMappings(t *testing.T) {
 	defer closer.Close()
 
 	in := fpMappings{
-		1: map[string]clientmodel.Fingerprint{
+		1: map[string]model.Fingerprint{
 			"foo": 1,
 			"bar": 2,
 		},
-		3: map[string]clientmodel.Fingerprint{
+		3: map[string]model.Fingerprint{
 			"baz": 4,
 		},
 	}
@@ -507,7 +586,7 @@ func TestCheckpointAndLoadFPMappings(t *testing.T) {
 	if err != nil {
 		t.Fatal(err)
 	}
-	if got, want := fp, clientmodel.Fingerprint(4); got != want {
+	if got, want := fp, model.Fingerprint(4); got != want {
 		t.Errorf("got highest FP %v, want %v", got, want)
 	}
 	if !reflect.DeepEqual(in, out) {
@@ -519,14 +598,14 @@ func testFingerprintsModifiedBefore(t *testing.T, encoding chunkEncoding) {
 	p, closer := newTestPersistence(t, encoding)
 	defer closer.Close()
 
-	m1 := clientmodel.Metric{"n1": "v1"}
-	m2 := clientmodel.Metric{"n2": "v2"}
-	m3 := clientmodel.Metric{"n1": "v2"}
+	m1 := model.Metric{"n1": "v1"}
+	m2 := model.Metric{"n2": "v2"}
+	m3 := model.Metric{"n1": "v2"}
 	p.archiveMetric(1, m1, 2, 4)
 	p.archiveMetric(2, m2, 1, 6)
 	p.archiveMetric(3, m3, 5, 5)
 
-	expectedFPs := map[clientmodel.Timestamp][]clientmodel.Fingerprint{
+	expectedFPs := map[model.Time][]model.Fingerprint{
 		0: {},
 		1: {},
 		2: {2},
@@ -546,25 +625,22 @@ func testFingerprintsModifiedBefore(t *testing.T, encoding chunkEncoding) {
 		}
 	}
 
-	unarchived, firstTime, err := p.unarchiveMetric(1)
+	unarchived, err := p.unarchiveMetric(1)
 	if err != nil {
 		t.Fatal(err)
 	}
 	if !unarchived {
-		t.Fatal("expected actual unarchival")
+		t.Error("expected actual unarchival")
 	}
-	if firstTime != 2 {
-		t.Errorf("expected first time 2, got %v", firstTime)
-	}
-	unarchived, firstTime, err = p.unarchiveMetric(1)
+	unarchived, err = p.unarchiveMetric(1)
 	if err != nil {
 		t.Fatal(err)
 	}
 	if unarchived {
-		t.Fatal("expected no unarchival")
+		t.Error("expected no unarchival")
 	}
 
-	expectedFPs = map[clientmodel.Timestamp][]clientmodel.Fingerprint{
+	expectedFPs = map[model.Time][]model.Fingerprint{
 		0: {},
 		1: {},
 		2: {2},
@@ -597,27 +673,27 @@ func testDropArchivedMetric(t *testing.T, encoding chunkEncoding) {
 	p, closer := newTestPersistence(t, encoding)
 	defer closer.Close()
 
-	m1 := clientmodel.Metric{"n1": "v1"}
-	m2 := clientmodel.Metric{"n2": "v2"}
+	m1 := model.Metric{"n1": "v1"}
+	m2 := model.Metric{"n2": "v2"}
 	p.archiveMetric(1, m1, 2, 4)
 	p.archiveMetric(2, m2, 1, 6)
 	p.indexMetric(1, m1)
 	p.indexMetric(2, m2)
 	p.waitForIndexing()
 
-	outFPs, err := p.fingerprintsForLabelPair(metric.LabelPair{Name: "n1", Value: "v1"})
+	outFPs, err := p.fingerprintsForLabelPair(model.LabelPair{Name: "n1", Value: "v1"})
 	if err != nil {
 		t.Fatal(err)
 	}
-	want := clientmodel.Fingerprints{1}
+	want := model.Fingerprints{1}
 	if !reflect.DeepEqual(outFPs, want) {
 		t.Errorf("want %#v, got %#v", want, outFPs)
 	}
-	outFPs, err = p.fingerprintsForLabelPair(metric.LabelPair{Name: "n2", Value: "v2"})
+	outFPs, err = p.fingerprintsForLabelPair(model.LabelPair{Name: "n2", Value: "v2"})
 	if err != nil {
 		t.Fatal(err)
 	}
-	want = clientmodel.Fingerprints{2}
+	want = model.Fingerprints{2}
 	if !reflect.DeepEqual(outFPs, want) {
 		t.Errorf("want %#v, got %#v", want, outFPs)
 	}
@@ -637,7 +713,7 @@ func testDropArchivedMetric(t *testing.T, encoding chunkEncoding) {
 	}
 	p.waitForIndexing()
 
-	outFPs, err = p.fingerprintsForLabelPair(metric.LabelPair{Name: "n1", Value: "v1"})
+	outFPs, err = p.fingerprintsForLabelPair(model.LabelPair{Name: "n1", Value: "v1"})
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -645,11 +721,11 @@ func testDropArchivedMetric(t *testing.T, encoding chunkEncoding) {
 	if !reflect.DeepEqual(outFPs, want) {
 		t.Errorf("want %#v, got %#v", want, outFPs)
 	}
-	outFPs, err = p.fingerprintsForLabelPair(metric.LabelPair{Name: "n2", Value: "v2"})
+	outFPs, err = p.fingerprintsForLabelPair(model.LabelPair{Name: "n2", Value: "v2"})
 	if err != nil {
 		t.Fatal(err)
 	}
-	want = clientmodel.Fingerprints{2}
+	want = model.Fingerprints{2}
 	if !reflect.DeepEqual(outFPs, want) {
 		t.Errorf("want %#v, got %#v", want, outFPs)
 	}
@@ -680,21 +756,21 @@ func testIndexing(t *testing.T, encoding chunkEncoding) {
 		{
 			fpToMetric: index.FingerprintMetricMapping{
 				0: {
-					clientmodel.MetricNameLabel: "metric_0",
-					"label_1":                   "value_1",
+					model.MetricNameLabel: "metric_0",
+					"label_1":             "value_1",
 				},
 				1: {
-					clientmodel.MetricNameLabel: "metric_0",
-					"label_2":                   "value_2",
-					"label_3":                   "value_3",
+					model.MetricNameLabel: "metric_0",
+					"label_2":             "value_2",
+					"label_3":             "value_3",
 				},
 				2: {
-					clientmodel.MetricNameLabel: "metric_1",
-					"label_1":                   "value_2",
+					model.MetricNameLabel: "metric_1",
+					"label_1":             "value_2",
 				},
 			},
 			expectedLnToLvs: index.LabelNameLabelValuesMapping{
-				clientmodel.MetricNameLabel: codable.LabelValueSet{
+				model.MetricNameLabel: codable.LabelValueSet{
 					"metric_0": struct{}{},
 					"metric_1": struct{}{},
 				},
@@ -710,27 +786,27 @@ func testIndexing(t *testing.T, encoding chunkEncoding) {
 				},
 			},
 			expectedLpToFps: index.LabelPairFingerprintsMapping{
-				metric.LabelPair{
-					Name:  clientmodel.MetricNameLabel,
+				model.LabelPair{
+					Name:  model.MetricNameLabel,
 					Value: "metric_0",
 				}: codable.FingerprintSet{0: struct{}{}, 1: struct{}{}},
-				metric.LabelPair{
-					Name:  clientmodel.MetricNameLabel,
+				model.LabelPair{
+					Name:  model.MetricNameLabel,
 					Value: "metric_1",
 				}: codable.FingerprintSet{2: struct{}{}},
-				metric.LabelPair{
+				model.LabelPair{
 					Name:  "label_1",
 					Value: "value_1",
 				}: codable.FingerprintSet{0: struct{}{}},
-				metric.LabelPair{
+				model.LabelPair{
 					Name:  "label_1",
 					Value: "value_2",
 				}: codable.FingerprintSet{2: struct{}{}},
-				metric.LabelPair{
+				model.LabelPair{
 					Name:  "label_2",
 					Value: "value_2",
 				}: codable.FingerprintSet{1: struct{}{}},
-				metric.LabelPair{
+				model.LabelPair{
 					Name:  "label_3",
 					Value: "value_3",
 				}: codable.FingerprintSet{1: struct{}{}},
@@ -738,21 +814,21 @@ func testIndexing(t *testing.T, encoding chunkEncoding) {
 		}, {
 			fpToMetric: index.FingerprintMetricMapping{
 				3: {
-					clientmodel.MetricNameLabel: "metric_0",
-					"label_1":                   "value_3",
+					model.MetricNameLabel: "metric_0",
+					"label_1":             "value_3",
 				},
 				4: {
-					clientmodel.MetricNameLabel: "metric_2",
-					"label_2":                   "value_2",
-					"label_3":                   "value_1",
+					model.MetricNameLabel: "metric_2",
+					"label_2":             "value_2",
+					"label_3":             "value_1",
 				},
 				5: {
-					clientmodel.MetricNameLabel: "metric_1",
-					"label_1":                   "value_3",
+					model.MetricNameLabel: "metric_1",
+					"label_1":             "value_3",
 				},
 			},
 			expectedLnToLvs: index.LabelNameLabelValuesMapping{
-				clientmodel.MetricNameLabel: codable.LabelValueSet{
+				model.MetricNameLabel: codable.LabelValueSet{
 					"metric_0": struct{}{},
 					"metric_1": struct{}{},
 					"metric_2": struct{}{},
@@ -771,39 +847,39 @@ func testIndexing(t *testing.T, encoding chunkEncoding) {
 				},
 			},
 			expectedLpToFps: index.LabelPairFingerprintsMapping{
-				metric.LabelPair{
-					Name:  clientmodel.MetricNameLabel,
+				model.LabelPair{
+					Name:  model.MetricNameLabel,
 					Value: "metric_0",
 				}: codable.FingerprintSet{0: struct{}{}, 1: struct{}{}, 3: struct{}{}},
-				metric.LabelPair{
-					Name:  clientmodel.MetricNameLabel,
+				model.LabelPair{
+					Name:  model.MetricNameLabel,
 					Value: "metric_1",
 				}: codable.FingerprintSet{2: struct{}{}, 5: struct{}{}},
-				metric.LabelPair{
-					Name:  clientmodel.MetricNameLabel,
+				model.LabelPair{
+					Name:  model.MetricNameLabel,
 					Value: "metric_2",
 				}: codable.FingerprintSet{4: struct{}{}},
-				metric.LabelPair{
+				model.LabelPair{
 					Name:  "label_1",
 					Value: "value_1",
 				}: codable.FingerprintSet{0: struct{}{}},
-				metric.LabelPair{
+				model.LabelPair{
 					Name:  "label_1",
 					Value: "value_2",
 				}: codable.FingerprintSet{2: struct{}{}},
-				metric.LabelPair{
+				model.LabelPair{
 					Name:  "label_1",
 					Value: "value_3",
 				}: codable.FingerprintSet{3: struct{}{}, 5: struct{}{}},
-				metric.LabelPair{
+				model.LabelPair{
 					Name:  "label_2",
 					Value: "value_2",
 				}: codable.FingerprintSet{1: struct{}{}, 4: struct{}{}},
-				metric.LabelPair{
+				model.LabelPair{
 					Name:  "label_3",
 					Value: "value_1",
 				}: codable.FingerprintSet{4: struct{}{}},
-				metric.LabelPair{
+				model.LabelPair{
 					Name:  "label_3",
 					Value: "value_3",
 				}: codable.FingerprintSet{1: struct{}{}},
@@ -831,16 +907,13 @@ func testIndexing(t *testing.T, encoding chunkEncoding) {
 		verifyIndexedState(i, t, batches[i], indexedFpsToMetrics, p)
 		for fp, m := range b.fpToMetric {
 			p.unindexMetric(fp, m)
-			unarchived, firstTime, err := p.unarchiveMetric(fp)
+			unarchived, err := p.unarchiveMetric(fp)
 			if err != nil {
 				t.Fatal(err)
 			}
 			if !unarchived {
 				t.Errorf("%d. metric not unarchived", i)
 			}
-			if firstTime != 1 {
-				t.Errorf("%d. expected firstTime=1, got %v", i, firstTime)
-			}
 			delete(indexedFpsToMetrics, fp)
 		}
 	}
@@ -933,10 +1006,10 @@ func BenchmarkLoadChunksSequentially(b *testing.B) {
 		sequentialIndexes[i] = i
 	}
 
-	var fp clientmodel.Fingerprint
+	var fp model.Fingerprint
 	for i := 0; i < b.N; i++ {
 		for _, s := range fpStrings {
-			fp.LoadFromString(s)
+			fp, _ = model.FingerprintFromString(s)
 			cds, err := p.loadChunks(fp, sequentialIndexes, 0)
 			if err != nil {
 				b.Error(err)
@@ -955,10 +1028,10 @@ func BenchmarkLoadChunksRandomly(b *testing.B) {
 	}
 	randomIndexes := []int{1, 5, 6, 8, 11, 14, 18, 23, 29, 33, 42, 46}
 
-	var fp clientmodel.Fingerprint
+	var fp model.Fingerprint
 	for i := 0; i < b.N; i++ {
 		for _, s := range fpStrings {
-			fp.LoadFromString(s)
+			fp, _ = model.FingerprintFromString(s)
 			cds, err := p.loadChunks(fp, randomIndexes, 0)
 			if err != nil {
 				b.Error(err)
@@ -975,11 +1048,11 @@ func BenchmarkLoadChunkDescs(b *testing.B) {
 		basePath: "fixtures",
 	}
 
-	var fp clientmodel.Fingerprint
+	var fp model.Fingerprint
 	for i := 0; i < b.N; i++ {
 		for _, s := range fpStrings {
-			fp.LoadFromString(s)
-			cds, err := p.loadChunkDescs(fp, clientmodel.Latest)
+			fp, _ = model.FingerprintFromString(s)
+			cds, err := p.loadChunkDescs(fp, 0)
 			if err != nil {
 				b.Error(err)
 			}
