@@ -18,7 +18,7 @@ import (
 	"sync"
 	"time"
 
-	clientmodel "github.com/prometheus/client_golang/model"
+	"github.com/prometheus/common/model"
 
 	"github.com/prometheus/prometheus/storage/metric"
 )
@@ -36,22 +36,22 @@ const (
 
 // fingerprintSeriesPair pairs a fingerprint with a memorySeries pointer.
 type fingerprintSeriesPair struct {
-	fp     clientmodel.Fingerprint
+	fp     model.Fingerprint
 	series *memorySeries
 }
 
 // seriesMap maps fingerprints to memory series. All its methods are
 // goroutine-safe. A SeriesMap is effectively is a goroutine-safe version of
-// map[clientmodel.Fingerprint]*memorySeries.
+// map[model.Fingerprint]*memorySeries.
 type seriesMap struct {
 	mtx sync.RWMutex
-	m   map[clientmodel.Fingerprint]*memorySeries
+	m   map[model.Fingerprint]*memorySeries
 }
 
 // newSeriesMap returns a newly allocated empty seriesMap. To create a seriesMap
 // based on a prefilled map, use an explicit initializer.
 func newSeriesMap() *seriesMap {
-	return &seriesMap{m: make(map[clientmodel.Fingerprint]*memorySeries)}
+	return &seriesMap{m: make(map[model.Fingerprint]*memorySeries)}
 }
 
 // length returns the number of mappings in the seriesMap.
@@ -64,7 +64,7 @@ func (sm *seriesMap) length() int {
 
 // get returns a memorySeries for a fingerprint. Return values have the same
 // semantics as the native Go map.
-func (sm *seriesMap) get(fp clientmodel.Fingerprint) (s *memorySeries, ok bool) {
+func (sm *seriesMap) get(fp model.Fingerprint) (s *memorySeries, ok bool) {
 	sm.mtx.RLock()
 	defer sm.mtx.RUnlock()
 
@@ -73,7 +73,7 @@ func (sm *seriesMap) get(fp clientmodel.Fingerprint) (s *memorySeries, ok bool)
 }
 
 // put adds a mapping to the seriesMap. It panics if s == nil.
-func (sm *seriesMap) put(fp clientmodel.Fingerprint, s *memorySeries) {
+func (sm *seriesMap) put(fp model.Fingerprint, s *memorySeries) {
 	sm.mtx.Lock()
 	defer sm.mtx.Unlock()
 
@@ -84,7 +84,7 @@ func (sm *seriesMap) put(fp clientmodel.Fingerprint, s *memorySeries) {
 }
 
 // del removes a mapping from the series Map.
-func (sm *seriesMap) del(fp clientmodel.Fingerprint) {
+func (sm *seriesMap) del(fp model.Fingerprint) {
 	sm.mtx.Lock()
 	defer sm.mtx.Unlock()
 
@@ -120,8 +120,8 @@ func (sm *seriesMap) iter() <-chan fingerprintSeriesPair {
 // for iterating over a map with a 'range' clause. However, if the next element
 // in iteration order is removed after the current element has been received
 // from the channel, it will still be produced by the channel.
-func (sm *seriesMap) fpIter() <-chan clientmodel.Fingerprint {
-	ch := make(chan clientmodel.Fingerprint)
+func (sm *seriesMap) fpIter() <-chan model.Fingerprint {
+	ch := make(chan model.Fingerprint)
 	go func() {
 		sm.mtx.RLock()
 		for fp := range sm.m {
@@ -136,7 +136,7 @@ func (sm *seriesMap) fpIter() <-chan clientmodel.Fingerprint {
 }
 
 type memorySeries struct {
-	metric clientmodel.Metric
+	metric model.Metric
 	// Sorted by start time, overlapping chunk ranges are forbidden.
 	chunkDescs []*chunkDesc
 	// The index (within chunkDescs above) of the first chunkDesc that
@@ -161,7 +161,10 @@ type memorySeries struct {
 	// chunkDescsOffset is not 0. It can be used to save the firstTime of the
 	// first chunk before its chunk desc is evicted. In doubt, this field is
 	// just set to the oldest possible timestamp.
-	savedFirstTime clientmodel.Timestamp
+	savedFirstTime model.Time
+	// The timestamp of the last sample in this series. Needed for fast access to
+	// ensure timestamp monotonicity during ingestion.
+	lastTime model.Time
 	// Whether the current head chunk has already been finished.  If true,
 	// the current head chunk must not be modified anymore.
 	headChunkClosed bool
@@ -175,35 +178,36 @@ type memorySeries struct {
 }
 
 // newMemorySeries returns a pointer to a newly allocated memorySeries for the
-// given metric. reallyNew defines if the memorySeries is a genuinely new series
-// or (if false) a series for a metric being unarchived, i.e. a series that
-// existed before but has been evicted from memory. If reallyNew is false,
-// firstTime is ignored (and set to the lowest possible timestamp instead - it
-// will be set properly upon the first eviction of chunkDescs).
-func newMemorySeries(
-	m clientmodel.Metric,
-	reallyNew bool,
-	firstTime clientmodel.Timestamp,
-) *memorySeries {
-	if !reallyNew {
-		firstTime = clientmodel.Earliest
+// given metric. chunkDescs and modTime in the new series are set according to
+// the provided parameters. chunkDescs can be nil or empty if this is a
+// genuinely new time series (i.e. not one that is being unarchived). In that
+// case, headChunkClosed is set to false, and firstTime and lastTime are both
+// set to model.Earliest. The zero value for modTime can be used if the
+// modification time of the series file is unknown (e.g. if this is a genuinely
+// new series).
+func newMemorySeries(m model.Metric, chunkDescs []*chunkDesc, modTime time.Time) *memorySeries {
+	firstTime := model.Earliest
+	lastTime := model.Earliest
+	if len(chunkDescs) > 0 {
+		firstTime = chunkDescs[0].firstTime()
+		lastTime = chunkDescs[len(chunkDescs)-1].lastTime()
 	}
-	s := memorySeries{
-		metric:          m,
-		headChunkClosed: !reallyNew,
-		savedFirstTime:  firstTime,
+	return &memorySeries{
+		metric:           m,
+		chunkDescs:       chunkDescs,
+		headChunkClosed:  len(chunkDescs) > 0,
+		savedFirstTime:   firstTime,
+		lastTime:         lastTime,
+		persistWatermark: len(chunkDescs),
+		modTime:          modTime,
 	}
-	if !reallyNew {
-		s.chunkDescsOffset = -1
-	}
-	return &s
 }
 
 // add adds a sample pair to the series. It returns the number of newly
 // completed chunks (which are now eligible for persistence).
 //
 // The caller must have locked the fingerprint of the series.
-func (s *memorySeries) add(v *metric.SamplePair) int {
+func (s *memorySeries) add(v *model.SamplePair) int {
 	if len(s.chunkDescs) == 0 || s.headChunkClosed {
 		newHead := newChunkDesc(newChunk())
 		s.chunkDescs = append(s.chunkDescs, newHead)
@@ -231,6 +235,8 @@ func (s *memorySeries) add(v *metric.SamplePair) int {
 	for _, c := range chunks[1:] {
 		s.chunkDescs = append(s.chunkDescs, newChunkDesc(c))
 	}
+
+	s.lastTime = v.Timestamp
 	return len(chunks) - 1
 }
 
@@ -243,7 +249,7 @@ func (s *memorySeries) maybeCloseHeadChunk() bool {
 	if s.headChunkClosed {
 		return false
 	}
-	if time.Now().Sub(s.head().lastTime().Time()) > headChunkTimeout {
+	if time.Now().Sub(s.lastTime.Time()) > headChunkTimeout {
 		s.headChunkClosed = true
 		// Since we cannot modify the head chunk from now on, we
 		// don't need to bother with cloning anymore.
@@ -275,7 +281,7 @@ func (s *memorySeries) evictChunkDescs(iOldestNotEvicted int) {
 
 // dropChunks removes chunkDescs older than t. The caller must have locked the
 // fingerprint of the series.
-func (s *memorySeries) dropChunks(t clientmodel.Timestamp) {
+func (s *memorySeries) dropChunks(t model.Time) {
 	keepIdx := len(s.chunkDescs)
 	for i, cd := range s.chunkDescs {
 		if !cd.lastTime().Before(t) {
@@ -302,7 +308,7 @@ func (s *memorySeries) dropChunks(t clientmodel.Timestamp) {
 
 // preloadChunks is an internal helper method.
 func (s *memorySeries) preloadChunks(
-	indexes []int, fp clientmodel.Fingerprint, mss *memorySeriesStorage,
+	indexes []int, fp model.Fingerprint, mss *memorySeriesStorage,
 ) ([]*chunkDesc, error) {
 	loadIndexes := []int{}
 	pinnedChunkDescs := make([]*chunkDesc, 0, len(indexes))
@@ -337,7 +343,7 @@ func (s *memorySeries) preloadChunks(
 }
 
 /*
-func (s *memorySeries) preloadChunksAtTime(t clientmodel.Timestamp, p *persistence) (chunkDescs, error) {
+func (s *memorySeries) preloadChunksAtTime(t model.Time, p *persistence) (chunkDescs, error) {
 	s.mtx.Lock()
 	defer s.mtx.Unlock()
 
@@ -370,15 +376,15 @@ func (s *memorySeries) preloadChunksAtTime(t clientmodel.Timestamp, p *persisten
 // preloadChunksForRange loads chunks for the given range from the persistence.
 // The caller must have locked the fingerprint of the series.
 func (s *memorySeries) preloadChunksForRange(
-	from clientmodel.Timestamp, through clientmodel.Timestamp,
-	fp clientmodel.Fingerprint, mss *memorySeriesStorage,
+	from model.Time, through model.Time,
+	fp model.Fingerprint, mss *memorySeriesStorage,
 ) ([]*chunkDesc, error) {
-	firstChunkDescTime := clientmodel.Latest
+	firstChunkDescTime := model.Latest
 	if len(s.chunkDescs) > 0 {
 		firstChunkDescTime = s.chunkDescs[0].firstTime()
 	}
 	if s.chunkDescsOffset != 0 && from.Before(firstChunkDescTime) {
-		cds, err := mss.loadChunkDescs(fp, firstChunkDescTime)
+		cds, err := mss.loadChunkDescs(fp, s.persistWatermark)
 		if err != nil {
 			return nil, err
 		}
@@ -441,7 +447,7 @@ func (s *memorySeries) head() *chunkDesc {
 
 // firstTime returns the timestamp of the first sample in the series. The caller
 // must have locked the fingerprint of the memorySeries.
-func (s *memorySeries) firstTime() clientmodel.Timestamp {
+func (s *memorySeries) firstTime() model.Time {
 	if s.chunkDescsOffset == 0 && len(s.chunkDescs) > 0 {
 		return s.chunkDescs[0].firstTime()
 	}
@@ -476,7 +482,7 @@ type memorySeriesIterator struct {
 }
 
 // ValueAtTime implements SeriesIterator.
-func (it *memorySeriesIterator) ValueAtTime(t clientmodel.Timestamp) metric.Values {
+func (it *memorySeriesIterator) ValueAtTime(t model.Time) []model.SamplePair {
 	// The most common case. We are iterating through a chunk.
 	if it.chunkIt != nil && it.chunkIt.contains(t) {
 		return it.chunkIt.valueAtTime(t)
@@ -491,7 +497,7 @@ func (it *memorySeriesIterator) ValueAtTime(t clientmodel.Timestamp) metric.Valu
 	ts := it.chunkIt.timestampAtIndex(0)
 	if !t.After(ts) {
 		// return first value of first chunk
-		return metric.Values{metric.SamplePair{
+		return []model.SamplePair{{
 			Timestamp: ts,
 			Value:     it.chunkIt.sampleValueAtIndex(0),
 		}}
@@ -502,7 +508,7 @@ func (it *memorySeriesIterator) ValueAtTime(t clientmodel.Timestamp) metric.Valu
 	ts = it.chunkIt.lastTimestamp()
 	if !t.Before(ts) {
 		// return last value of last chunk
-		return metric.Values{metric.SamplePair{
+		return []model.SamplePair{{
 			Timestamp: ts,
 			Value:     it.chunkIt.sampleValueAtIndex(it.chunkIt.length() - 1),
 		}}
@@ -520,14 +526,14 @@ func (it *memorySeriesIterator) ValueAtTime(t clientmodel.Timestamp) metric.Valu
 	ts = it.chunkIt.lastTimestamp()
 	if t.After(ts) {
 		// We ended up between two chunks.
-		sp1 := metric.SamplePair{
+		sp1 := model.SamplePair{
 			Timestamp: ts,
 			Value:     it.chunkIt.sampleValueAtIndex(it.chunkIt.length() - 1),
 		}
 		it.chunkIt = it.chunkIterator(l - i + 1)
-		return metric.Values{
+		return []model.SamplePair{
 			sp1,
-			metric.SamplePair{
+			{
 				Timestamp: it.chunkIt.timestampAtIndex(0),
 				Value:     it.chunkIt.sampleValueAtIndex(0),
 			},
@@ -537,7 +543,7 @@ func (it *memorySeriesIterator) ValueAtTime(t clientmodel.Timestamp) metric.Valu
 }
 
 // BoundaryValues implements SeriesIterator.
-func (it *memorySeriesIterator) BoundaryValues(in metric.Interval) metric.Values {
+func (it *memorySeriesIterator) BoundaryValues(in metric.Interval) []model.SamplePair {
 	// Find the first chunk for which the first sample is within the interval.
 	i := sort.Search(len(it.chunks), func(i int) bool {
 		return !it.chunks[i].firstTime().Before(in.OldestInclusive)
@@ -548,7 +554,7 @@ func (it *memorySeriesIterator) BoundaryValues(in metric.Interval) metric.Values
 		i--
 	}
 
-	values := make(metric.Values, 0, 2)
+	values := make([]model.SamplePair, 0, 2)
 	for j, c := range it.chunks[i:] {
 		if c.firstTime().After(in.NewestInclusive) {
 			if len(values) == 1 {
@@ -557,7 +563,7 @@ func (it *memorySeriesIterator) BoundaryValues(in metric.Interval) metric.Values
 				// want must be the last value of the previous
 				// chunk. So backtrack...
 				chunkIt := it.chunkIterator(i + j - 1)
-				values = append(values, metric.SamplePair{
+				values = append(values, model.SamplePair{
 					Timestamp: chunkIt.lastTimestamp(),
 					Value:     chunkIt.lastSampleValue(),
 				})
@@ -584,7 +590,7 @@ func (it *memorySeriesIterator) BoundaryValues(in metric.Interval) metric.Values
 	if len(values) == 1 {
 		// We found exactly one value. In that case, add the most recent we know.
 		chunkIt := it.chunkIterator(len(it.chunks) - 1)
-		values = append(values, metric.SamplePair{
+		values = append(values, model.SamplePair{
 			Timestamp: chunkIt.lastTimestamp(),
 			Value:     chunkIt.lastSampleValue(),
 		})
@@ -596,7 +602,7 @@ func (it *memorySeriesIterator) BoundaryValues(in metric.Interval) metric.Values
 }
 
 // RangeValues implements SeriesIterator.
-func (it *memorySeriesIterator) RangeValues(in metric.Interval) metric.Values {
+func (it *memorySeriesIterator) RangeValues(in metric.Interval) []model.SamplePair {
 	// Find the first chunk for which the first sample is within the interval.
 	i := sort.Search(len(it.chunks), func(i int) bool {
 		return !it.chunks[i].firstTime().Before(in.OldestInclusive)
@@ -607,7 +613,7 @@ func (it *memorySeriesIterator) RangeValues(in metric.Interval) metric.Values {
 		i--
 	}
 
-	values := metric.Values{}
+	values := []model.SamplePair{}
 	for j, c := range it.chunks[i:] {
 		if c.firstTime().After(in.NewestInclusive) {
 			break
@@ -632,16 +638,16 @@ func (it *memorySeriesIterator) chunkIterator(i int) chunkIterator {
 type nopSeriesIterator struct{}
 
 // ValueAtTime implements SeriesIterator.
-func (_ nopSeriesIterator) ValueAtTime(t clientmodel.Timestamp) metric.Values {
-	return metric.Values{}
+func (i nopSeriesIterator) ValueAtTime(t model.Time) []model.SamplePair {
+	return []model.SamplePair{}
 }
 
 // BoundaryValues implements SeriesIterator.
-func (_ nopSeriesIterator) BoundaryValues(in metric.Interval) metric.Values {
-	return metric.Values{}
+func (i nopSeriesIterator) BoundaryValues(in metric.Interval) []model.SamplePair {
+	return []model.SamplePair{}
 }
 
 // RangeValues implements SeriesIterator.
-func (_ nopSeriesIterator) RangeValues(in metric.Interval) metric.Values {
-	return metric.Values{}
+func (i nopSeriesIterator) RangeValues(in metric.Interval) []model.SamplePair {
+	return []model.SamplePair{}
 }
