@@ -16,13 +16,13 @@ package local
 
 import (
 	"container/list"
+	"fmt"
 	"sync/atomic"
 	"time"
 
 	"github.com/prometheus/client_golang/prometheus"
-	"github.com/prometheus/log"
-
-	clientmodel "github.com/prometheus/client_golang/model"
+	"github.com/prometheus/common/log"
+	"github.com/prometheus/common/model"
 
 	"github.com/prometheus/prometheus/storage/metric"
 )
@@ -66,6 +66,34 @@ type evictRequest struct {
 // SyncStrategy is an enum to select a sync strategy for series files.
 type SyncStrategy int
 
+// String implements flag.Value.
+func (ss SyncStrategy) String() string {
+	switch ss {
+	case Adaptive:
+		return "adaptive"
+	case Always:
+		return "always"
+	case Never:
+		return "never"
+	}
+	return "<unknown>"
+}
+
+// Set implements flag.Value.
+func (ss *SyncStrategy) Set(s string) error {
+	switch s {
+	case "adaptive":
+		*ss = Adaptive
+	case "always":
+		*ss = Always
+	case "never":
+		*ss = Never
+	default:
+		return fmt.Errorf("invalid sync strategy: %s", s)
+	}
+	return nil
+}
+
 // Possible values for SyncStrategy.
 const (
 	_ SyncStrategy = iota
@@ -106,6 +134,7 @@ type memorySeriesStorage struct {
 	numSeries                   prometheus.Gauge
 	seriesOps                   *prometheus.CounterVec
 	ingestedSamplesCount        prometheus.Counter
+	outOfOrderSamplesCount      prometheus.Counter
 	invalidPreloadRequestsCount prometheus.Counter
 	maintainSeriesDuration      *prometheus.SummaryVec
 }
@@ -123,6 +152,7 @@ type MemorySeriesStorageOptions struct {
 	Dirty                      bool          // Force the storage to consider itself dirty on startup.
 	PedanticChecks             bool          // If dirty, perform crash-recovery checks on each series file.
 	SyncStrategy               SyncStrategy  // Which sync strategy to apply to series files.
+	MinShrinkRatio             float64       // Minimum ratio a series file has to shrink during truncation.
 }
 
 // NewMemorySeriesStorage returns a newly allocated Storage. Storage.Serve still
@@ -174,6 +204,12 @@ func NewMemorySeriesStorage(o *MemorySeriesStorageOptions) Storage {
 			Name:      "ingested_samples_total",
 			Help:      "The total number of samples ingested.",
 		}),
+		outOfOrderSamplesCount: prometheus.NewCounter(prometheus.CounterOpts{
+			Namespace: namespace,
+			Subsystem: subsystem,
+			Name:      "out_of_order_samples_total",
+			Help:      "The total number of samples that were discarded because their timestamps were at or before the last received sample for a series.",
+		}),
 		invalidPreloadRequestsCount: prometheus.NewCounter(prometheus.CounterOpts{
 			Namespace: namespace,
 			Subsystem: subsystem,
@@ -208,7 +244,12 @@ func (s *memorySeriesStorage) Start() (err error) {
 	}
 
 	var p *persistence
-	p, err = newPersistence(s.options.PersistenceStoragePath, s.options.Dirty, s.options.PedanticChecks, syncStrategy)
+	p, err = newPersistence(
+		s.options.PersistenceStoragePath,
+		s.options.Dirty, s.options.PedanticChecks,
+		syncStrategy,
+		s.options.MinShrinkRatio,
+	)
 	if err != nil {
 		return err
 	}
@@ -272,8 +313,8 @@ func (s *memorySeriesStorage) WaitForIndexing() {
 	s.persistence.waitForIndexing()
 }
 
-// NewIterator implements storage.
-func (s *memorySeriesStorage) NewIterator(fp clientmodel.Fingerprint) SeriesIterator {
+// NewIterator implements Storage.
+func (s *memorySeriesStorage) NewIterator(fp model.Fingerprint) SeriesIterator {
 	s.fpLocker.Lock(fp)
 	defer s.fpLocker.Unlock(fp)
 
@@ -288,29 +329,41 @@ func (s *memorySeriesStorage) NewIterator(fp clientmodel.Fingerprint) SeriesIter
 	}
 	return &boundedIterator{
 		it:    series.newIterator(),
-		start: clientmodel.Now().Add(-s.dropAfter),
+		start: model.Now().Add(-s.dropAfter),
 	}
 }
 
+// LastSampleForFingerprint implements Storage.
+func (s *memorySeriesStorage) LastSamplePairForFingerprint(fp model.Fingerprint) *model.SamplePair {
+	s.fpLocker.Lock(fp)
+	defer s.fpLocker.Unlock(fp)
+
+	series, ok := s.fpToSeries.get(fp)
+	if !ok {
+		return nil
+	}
+	return series.head().lastSamplePair()
+}
+
 // boundedIterator wraps a SeriesIterator and does not allow fetching
 // data from earlier than the configured start time.
 type boundedIterator struct {
 	it    SeriesIterator
-	start clientmodel.Timestamp
+	start model.Time
 }
 
 // ValueAtTime implements the SeriesIterator interface.
-func (bit *boundedIterator) ValueAtTime(ts clientmodel.Timestamp) metric.Values {
+func (bit *boundedIterator) ValueAtTime(ts model.Time) []model.SamplePair {
 	if ts < bit.start {
-		return metric.Values{}
+		return []model.SamplePair{}
 	}
 	return bit.it.ValueAtTime(ts)
 }
 
 // BoundaryValues implements the SeriesIterator interface.
-func (bit *boundedIterator) BoundaryValues(interval metric.Interval) metric.Values {
+func (bit *boundedIterator) BoundaryValues(interval metric.Interval) []model.SamplePair {
 	if interval.NewestInclusive < bit.start {
-		return metric.Values{}
+		return []model.SamplePair{}
 	}
 	if interval.OldestInclusive < bit.start {
 		interval.OldestInclusive = bit.start
@@ -319,9 +372,9 @@ func (bit *boundedIterator) BoundaryValues(interval metric.Interval) metric.Valu
 }
 
 // RangeValues implements the SeriesIterator interface.
-func (bit *boundedIterator) RangeValues(interval metric.Interval) metric.Values {
+func (bit *boundedIterator) RangeValues(interval metric.Interval) []model.SamplePair {
 	if interval.NewestInclusive < bit.start {
-		return metric.Values{}
+		return []model.SamplePair{}
 	}
 	if interval.OldestInclusive < bit.start {
 		interval.OldestInclusive = bit.start
@@ -336,54 +389,22 @@ func (s *memorySeriesStorage) NewPreloader() Preloader {
 	}
 }
 
-// FingerprintsForLabelMatchers implements Storage.
-func (s *memorySeriesStorage) FingerprintsForLabelMatchers(labelMatchers metric.LabelMatchers) clientmodel.Fingerprints {
-	var result map[clientmodel.Fingerprint]struct{}
-	for _, matcher := range labelMatchers {
-		intersection := map[clientmodel.Fingerprint]struct{}{}
-		switch matcher.Type {
-		case metric.Equal:
-			fps, err := s.persistence.fingerprintsForLabelPair(
-				metric.LabelPair{
-					Name:  matcher.Name,
-					Value: matcher.Value,
-				},
-			)
-			if err != nil {
-				log.Error("Error getting fingerprints for label pair: ", err)
-			}
-			if len(fps) == 0 {
-				return nil
-			}
-			for _, fp := range fps {
-				if _, ok := result[fp]; ok || result == nil {
-					intersection[fp] = struct{}{}
-				}
-			}
-		default:
-			values, err := s.persistence.labelValuesForLabelName(matcher.Name)
-			if err != nil {
-				log.Errorf("Error getting label values for label name %q: %v", matcher.Name, err)
-			}
-			matches := matcher.Filter(values)
-			if len(matches) == 0 {
-				return nil
-			}
-			for _, v := range matches {
-				fps, err := s.persistence.fingerprintsForLabelPair(
-					metric.LabelPair{
-						Name:  matcher.Name,
-						Value: v,
-					},
-				)
-				if err != nil {
-					log.Error("Error getting fingerprints for label pair: ", err)
-				}
-				for _, fp := range fps {
-					if _, ok := result[fp]; ok || result == nil {
-						intersection[fp] = struct{}{}
-					}
-				}
+// fingerprintsForLabelPairs returns the set of fingerprints that have the given labels.
+// This does not work with empty label values.
+func (s *memorySeriesStorage) fingerprintsForLabelPairs(pairs ...model.LabelPair) map[model.Fingerprint]struct{} {
+	var result map[model.Fingerprint]struct{}
+	for _, pair := range pairs {
+		intersection := map[model.Fingerprint]struct{}{}
+		fps, err := s.persistence.fingerprintsForLabelPair(pair)
+		if err != nil {
+			log.Error("Error getting fingerprints for label pair: ", err)
+		}
+		if len(fps) == 0 {
+			return nil
+		}
+		for _, fp := range fps {
+			if _, ok := result[fp]; ok || result == nil {
+				intersection[fp] = struct{}{}
 			}
 		}
 		if len(intersection) == 0 {
@@ -391,16 +412,78 @@ func (s *memorySeriesStorage) FingerprintsForLabelMatchers(labelMatchers metric.
 		}
 		result = intersection
 	}
+	return result
+}
 
-	fps := make(clientmodel.Fingerprints, 0, len(result))
-	for fp := range result {
-		fps = append(fps, fp)
+// MetricsForLabelMatchers implements Storage.
+func (s *memorySeriesStorage) MetricsForLabelMatchers(matchers ...*metric.LabelMatcher) map[model.Fingerprint]metric.Metric {
+	var (
+		equals  []model.LabelPair
+		filters []*metric.LabelMatcher
+	)
+	for _, lm := range matchers {
+		if lm.Type == metric.Equal && lm.Value != "" {
+			equals = append(equals, model.LabelPair{
+				Name:  lm.Name,
+				Value: lm.Value,
+			})
+		} else {
+			filters = append(filters, lm)
+		}
 	}
-	return fps
+
+	var resFPs map[model.Fingerprint]struct{}
+	if len(equals) > 0 {
+		resFPs = s.fingerprintsForLabelPairs(equals...)
+	} else {
+		// If we cannot make a preselection based on equality matchers, expanding the other matchers to labels
+		// and intersecting their fingerprints is still likely to be the best choice.
+		var remaining metric.LabelMatchers
+		for _, matcher := range filters {
+			// Equal matches are all empty values.
+			if matcher.Match("") {
+				remaining = append(remaining, matcher)
+				continue
+			}
+			intersection := map[model.Fingerprint]struct{}{}
+
+			matches := matcher.Filter(s.LabelValuesForLabelName(matcher.Name))
+			if len(matches) == 0 {
+				return nil
+			}
+			for _, v := range matches {
+				fps := s.fingerprintsForLabelPairs(model.LabelPair{
+					Name:  matcher.Name,
+					Value: v,
+				})
+				for fp := range fps {
+					if _, ok := resFPs[fp]; ok || resFPs == nil {
+						intersection[fp] = struct{}{}
+					}
+				}
+			}
+			resFPs = intersection
+		}
+		// The intersected matchers no longer need to be compared against the actual metrics.
+		filters = remaining
+	}
+
+	result := make(map[model.Fingerprint]metric.Metric, len(resFPs))
+	for fp := range resFPs {
+		result[fp] = s.MetricForFingerprint(fp)
+	}
+	for _, matcher := range filters {
+		for fp, met := range result {
+			if !matcher.Match(met.Metric[matcher.Name]) {
+				delete(result, fp)
+			}
+		}
+	}
+	return result
 }
 
 // LabelValuesForLabelName implements Storage.
-func (s *memorySeriesStorage) LabelValuesForLabelName(labelName clientmodel.LabelName) clientmodel.LabelValues {
+func (s *memorySeriesStorage) LabelValuesForLabelName(labelName model.LabelName) model.LabelValues {
 	lvs, err := s.persistence.labelValuesForLabelName(labelName)
 	if err != nil {
 		log.Errorf("Error getting label values for label name %q: %v", labelName, err)
@@ -409,7 +492,7 @@ func (s *memorySeriesStorage) LabelValuesForLabelName(labelName clientmodel.Labe
 }
 
 // MetricForFingerprint implements Storage.
-func (s *memorySeriesStorage) MetricForFingerprint(fp clientmodel.Fingerprint) clientmodel.COWMetric {
+func (s *memorySeriesStorage) MetricForFingerprint(fp model.Fingerprint) metric.Metric {
 	s.fpLocker.Lock(fp)
 	defer s.fpLocker.Unlock(fp)
 
@@ -417,21 +500,50 @@ func (s *memorySeriesStorage) MetricForFingerprint(fp clientmodel.Fingerprint) c
 	if ok {
 		// Wrap the returned metric in a copy-on-write (COW) metric here because
 		// the caller might mutate it.
-		return clientmodel.COWMetric{
+		return metric.Metric{
 			Metric: series.metric,
 		}
 	}
-	metric, err := s.persistence.archivedMetric(fp)
+	met, err := s.persistence.archivedMetric(fp)
 	if err != nil {
 		log.Errorf("Error retrieving archived metric for fingerprint %v: %v", fp, err)
 	}
-	return clientmodel.COWMetric{
-		Metric: metric,
+
+	return metric.Metric{
+		Metric: met,
+		Copied: false,
+	}
+}
+
+// DropMetric implements Storage.
+func (s *memorySeriesStorage) DropMetricsForFingerprints(fps ...model.Fingerprint) {
+	for _, fp := range fps {
+		s.fpLocker.Lock(fp)
+
+		if series, ok := s.fpToSeries.get(fp); ok {
+			s.fpToSeries.del(fp)
+			s.numSeries.Dec()
+			s.persistence.unindexMetric(fp, series.metric)
+		} else if err := s.persistence.purgeArchivedMetric(fp); err != nil {
+			log.Errorf("Error purging metric with fingerprint %v: %v", fp, err)
+		}
+		// Attempt to delete series file in any case.
+		if _, err := s.persistence.deleteSeriesFile(fp); err != nil {
+			log.Errorf("Error deleting series file for %v: %v", fp, err)
+		}
+
+		s.fpLocker.Unlock(fp)
+		s.seriesOps.WithLabelValues(requestedPurge).Inc()
 	}
 }
 
 // Append implements Storage.
-func (s *memorySeriesStorage) Append(sample *clientmodel.Sample) {
+func (s *memorySeriesStorage) Append(sample *model.Sample) {
+	for ln, lv := range sample.Metric {
+		if len(lv) == 0 {
+			delete(sample.Metric, ln)
+		}
+	}
 	if s.getNumChunksToPersist() >= s.maxChunksToPersist {
 		log.Warnf(
 			"%d chunks waiting for persistence, sample ingestion suspended.",
@@ -455,7 +567,20 @@ func (s *memorySeriesStorage) Append(sample *clientmodel.Sample) {
 		s.fpLocker.Lock(fp)
 	}
 	series := s.getOrCreateSeries(fp, sample.Metric)
-	completedChunksCount := series.add(&metric.SamplePair{
+
+	if sample.Timestamp <= series.lastTime {
+		// Don't log and track equal timestamps, as they are a common occurrence
+		// when using client-side timestamps (e.g. Pushgateway or federation).
+		// It would be even better to also compare the sample values here, but
+		// we don't have efficient access to a series's last value.
+		if sample.Timestamp != series.lastTime {
+			log.Warnf("Ignoring sample with out-of-order timestamp for fingerprint %v (%v): %v is not after %v", fp, series.metric, sample.Timestamp, series.lastTime)
+			s.outOfOrderSamplesCount.Inc()
+		}
+		s.fpLocker.Unlock(fp)
+		return
+	}
+	completedChunksCount := series.add(&model.SamplePair{
 		Value:     sample.Value,
 		Timestamp: sample.Timestamp,
 	})
@@ -464,21 +589,33 @@ func (s *memorySeriesStorage) Append(sample *clientmodel.Sample) {
 	s.incNumChunksToPersist(completedChunksCount)
 }
 
-func (s *memorySeriesStorage) getOrCreateSeries(fp clientmodel.Fingerprint, m clientmodel.Metric) *memorySeries {
+func (s *memorySeriesStorage) getOrCreateSeries(fp model.Fingerprint, m model.Metric) *memorySeries {
 	series, ok := s.fpToSeries.get(fp)
 	if !ok {
-		unarchived, firstTime, err := s.persistence.unarchiveMetric(fp)
+		var cds []*chunkDesc
+		var modTime time.Time
+		unarchived, err := s.persistence.unarchiveMetric(fp)
 		if err != nil {
-			log.Errorf("Error unarchiving fingerprint %v: %v", fp, err)
+			log.Errorf("Error unarchiving fingerprint %v (metric %v): %v", fp, m, err)
 		}
 		if unarchived {
 			s.seriesOps.WithLabelValues(unarchive).Inc()
+			// We have to load chunkDescs anyway to do anything with
+			// the series, so let's do it right now so that we don't
+			// end up with a series without any chunkDescs for a
+			// while (which is confusing as it makes the series
+			// appear as archived or purged).
+			cds, err = s.loadChunkDescs(fp, 0)
+			if err != nil {
+				log.Errorf("Error loading chunk descs for fingerprint %v (metric %v): %v", fp, m, err)
+			}
+			modTime = s.persistence.seriesFileModTime(fp)
 		} else {
 			// This was a genuinely new series, so index the metric.
 			s.persistence.indexMetric(fp, m)
 			s.seriesOps.WithLabelValues(create).Inc()
 		}
-		series = newMemorySeries(m, !unarchived, firstTime)
+		series = newMemorySeries(m, cds, modTime)
 		s.fpToSeries.put(fp, series)
 		s.numSeries.Inc()
 	}
@@ -486,8 +623,8 @@ func (s *memorySeriesStorage) getOrCreateSeries(fp clientmodel.Fingerprint, m cl
 }
 
 func (s *memorySeriesStorage) preloadChunksForRange(
-	fp clientmodel.Fingerprint,
-	from clientmodel.Timestamp, through clientmodel.Timestamp,
+	fp model.Fingerprint,
+	from model.Time, through model.Time,
 	stalenessDelta time.Duration,
 ) ([]*chunkDesc, error) {
 	s.fpLocker.Lock(fp)
@@ -640,10 +777,10 @@ func (s *memorySeriesStorage) waitForNextFP(numberOfFPs int, maxWaitDurationFact
 // cycleThroughMemoryFingerprints returns a channel that emits fingerprints for
 // series in memory in a throttled fashion. It continues to cycle through all
 // fingerprints in memory until s.loopStopping is closed.
-func (s *memorySeriesStorage) cycleThroughMemoryFingerprints() chan clientmodel.Fingerprint {
-	memoryFingerprints := make(chan clientmodel.Fingerprint)
+func (s *memorySeriesStorage) cycleThroughMemoryFingerprints() chan model.Fingerprint {
+	memoryFingerprints := make(chan model.Fingerprint)
 	go func() {
-		var fpIter <-chan clientmodel.Fingerprint
+		var fpIter <-chan model.Fingerprint
 
 		defer func() {
 			if fpIter != nil {
@@ -687,14 +824,14 @@ func (s *memorySeriesStorage) cycleThroughMemoryFingerprints() chan clientmodel.
 // cycleThroughArchivedFingerprints returns a channel that emits fingerprints
 // for archived series in a throttled fashion. It continues to cycle through all
 // archived fingerprints until s.loopStopping is closed.
-func (s *memorySeriesStorage) cycleThroughArchivedFingerprints() chan clientmodel.Fingerprint {
-	archivedFingerprints := make(chan clientmodel.Fingerprint)
+func (s *memorySeriesStorage) cycleThroughArchivedFingerprints() chan model.Fingerprint {
+	archivedFingerprints := make(chan model.Fingerprint)
 	go func() {
 		defer close(archivedFingerprints)
 
 		for {
 			archivedFPs, err := s.persistence.fingerprintsModifiedBefore(
-				clientmodel.TimestampFromTime(time.Now()).Add(-s.dropAfter),
+				model.Now().Add(-s.dropAfter),
 			)
 			if err != nil {
 				log.Error("Failed to lookup archived fingerprint ranges: ", err)
@@ -746,11 +883,15 @@ loop:
 		case <-s.loopStopping:
 			break loop
 		case <-checkpointTimer.C:
-			s.persistence.checkpointSeriesMapAndHeads(s.fpToSeries, s.fpLocker)
-			dirtySeriesCount = 0
+			err := s.persistence.checkpointSeriesMapAndHeads(s.fpToSeries, s.fpLocker)
+			if err != nil {
+				log.Errorln("Error while checkpointing:", err)
+			} else {
+				dirtySeriesCount = 0
+			}
 			checkpointTimer.Reset(s.checkpointInterval)
 		case fp := <-memoryFingerprints:
-			if s.maintainMemorySeries(fp, clientmodel.TimestampFromTime(time.Now()).Add(-s.dropAfter)) {
+			if s.maintainMemorySeries(fp, model.Now().Add(-s.dropAfter)) {
 				dirtySeriesCount++
 				// Check if we have enough "dirty" series so that we need an early checkpoint.
 				// However, if we are already behind persisting chunks, creating a checkpoint
@@ -758,13 +899,13 @@ loop:
 				// while in a situation like that, where we are clearly lacking speed of disk
 				// maintenance, the best we can do for crash recovery is to persist chunks as
 				// quickly as possible. So only checkpoint if the storage is not in "graceful
-				// degratadion mode".
+				// degradation mode".
 				if dirtySeriesCount >= s.checkpointDirtySeriesLimit && !s.isDegraded() {
 					checkpointTimer.Reset(0)
 				}
 			}
 		case fp := <-archivedFingerprints:
-			s.maintainArchivedSeries(fp, clientmodel.TimestampFromTime(time.Now()).Add(-s.dropAfter))
+			s.maintainArchivedSeries(fp, model.Now().Add(-s.dropAfter))
 		}
 	}
 	// Wait until both channels are closed.
@@ -806,7 +947,7 @@ loop:
 //
 // Finally, it evicts chunkDescs if there are too many.
 func (s *memorySeriesStorage) maintainMemorySeries(
-	fp clientmodel.Fingerprint, beforeTime clientmodel.Timestamp,
+	fp model.Fingerprint, beforeTime model.Time,
 ) (becameDirty bool) {
 	defer func(begin time.Time) {
 		s.maintainSeriesDuration.WithLabelValues(maintainInMemory).Observe(
@@ -848,21 +989,8 @@ func (s *memorySeriesStorage) maintainMemorySeries(
 	if iOldestNotEvicted == -1 {
 		s.fpToSeries.del(fp)
 		s.numSeries.Dec()
-		// Make sure we have a head chunk descriptor (a freshly
-		// unarchived series has none).
-		if len(series.chunkDescs) == 0 {
-			cds, err := s.loadChunkDescs(fp, clientmodel.Latest)
-			if err != nil {
-				log.Errorf(
-					"Could not load chunk descriptors prior to archiving metric %v, metric will not be archived: %v",
-					series.metric, err,
-				)
-				return
-			}
-			series.chunkDescs = cds
-		}
 		if err := s.persistence.archiveMetric(
-			fp, series.metric, series.firstTime(), series.head().lastTime(),
+			fp, series.metric, series.firstTime(), series.lastTime,
 		); err != nil {
 			log.Errorf("Error archiving metric %v: %v", series.metric, err)
 			return
@@ -871,7 +999,7 @@ func (s *memorySeriesStorage) maintainMemorySeries(
 		return
 	}
 	// If we are here, the series is not archived, so check for chunkDesc
-	// eviction next
+	// eviction next.
 	series.evictChunkDescs(iOldestNotEvicted)
 
 	return series.dirty && !seriesWasDirty
@@ -887,7 +1015,7 @@ func (s *memorySeriesStorage) maintainMemorySeries(
 //
 // The caller must have locked the fp.
 func (s *memorySeriesStorage) writeMemorySeries(
-	fp clientmodel.Fingerprint, series *memorySeries, beforeTime clientmodel.Timestamp,
+	fp model.Fingerprint, series *memorySeries, beforeTime model.Time,
 ) bool {
 	cds := series.chunksToPersist()
 	defer func() {
@@ -956,7 +1084,7 @@ func (s *memorySeriesStorage) writeMemorySeries(
 
 // maintainArchivedSeries drops chunks older than beforeTime from an archived
 // series. If the series contains no chunks after that, it is purged entirely.
-func (s *memorySeriesStorage) maintainArchivedSeries(fp clientmodel.Fingerprint, beforeTime clientmodel.Timestamp) {
+func (s *memorySeriesStorage) maintainArchivedSeries(fp model.Fingerprint, beforeTime model.Time) {
 	defer func(begin time.Time) {
 		s.maintainSeriesDuration.WithLabelValues(maintainArchived).Observe(
 			float64(time.Since(begin)) / float64(time.Millisecond),
@@ -990,17 +1118,19 @@ func (s *memorySeriesStorage) maintainArchivedSeries(fp clientmodel.Fingerprint,
 		s.seriesOps.WithLabelValues(archivePurge).Inc()
 		return
 	}
-	s.persistence.updateArchivedTimeRange(fp, newFirstTime, lastTime)
+	if err := s.persistence.updateArchivedTimeRange(fp, newFirstTime, lastTime); err != nil {
+		log.Errorf("Error updating archived time range for fingerprint %v: %s", fp, err)
+	}
 }
 
 // See persistence.loadChunks for detailed explanation.
-func (s *memorySeriesStorage) loadChunks(fp clientmodel.Fingerprint, indexes []int, indexOffset int) ([]chunk, error) {
+func (s *memorySeriesStorage) loadChunks(fp model.Fingerprint, indexes []int, indexOffset int) ([]chunk, error) {
 	return s.persistence.loadChunks(fp, indexes, indexOffset)
 }
 
 // See persistence.loadChunkDescs for detailed explanation.
-func (s *memorySeriesStorage) loadChunkDescs(fp clientmodel.Fingerprint, beforeTime clientmodel.Timestamp) ([]*chunkDesc, error) {
-	return s.persistence.loadChunkDescs(fp, beforeTime)
+func (s *memorySeriesStorage) loadChunkDescs(fp model.Fingerprint, offsetFromEnd int) ([]*chunkDesc, error) {
+	return s.persistence.loadChunkDescs(fp, offsetFromEnd)
 }
 
 // getNumChunksToPersist returns numChunksToPersist in a goroutine-safe way.
@@ -1059,6 +1189,7 @@ func (s *memorySeriesStorage) Describe(ch chan<- *prometheus.Desc) {
 	ch <- s.numSeries.Desc()
 	s.seriesOps.Describe(ch)
 	ch <- s.ingestedSamplesCount.Desc()
+	ch <- s.outOfOrderSamplesCount.Desc()
 	ch <- s.invalidPreloadRequestsCount.Desc()
 	ch <- numMemChunksDesc
 	s.maintainSeriesDuration.Describe(ch)
@@ -1083,6 +1214,7 @@ func (s *memorySeriesStorage) Collect(ch chan<- prometheus.Metric) {
 	ch <- s.numSeries
 	s.seriesOps.Collect(ch)
 	ch <- s.ingestedSamplesCount
+	ch <- s.outOfOrderSamplesCount
 	ch <- s.invalidPreloadRequestsCount
 	ch <- prometheus.MustNewConstMetric(
 		numMemChunksDesc,
